#include <cstdio>
#include <iostream>
#include "phi.hpp"
#include <sdsl/rmq_support.hpp>
#include <sdsl/construct.hpp>

using namespace sdsl;
using namespace std;


uint64_t repeats_count = 0;

// clear the space used by x
template<class T>
void clear(T& x)
{
	T tmp;
	tmp.swap(x);
}

// Computation of CT
template<class lcp_t, class bwt_t>
vector<uint32_t> ct1(lcp_t& lcp, bwt_t& bwt, uint64_t largest_symbol)
{
	rmq_succinct_sct<> rmq(&lcp);
	vector<uint32_t> ct(bwt.size(), 0);
	vector<uint64_t> last(largest_symbol+1, 0);
	for(uint64_t i=0; i<ct.size(); ++i)
	{
		uint64_t c = bwt[i];
		if(last[c] != 0)
		{
			uint64_t j = rmq(last[c],i);
			++ct[j];
		}
		last[c] = i+1;
	}
	for(uint64_t i=1; i<ct.size(); ++i)
	{
		ct[i] += ct[i-1];
	}
	return ct;
}

// Computation of CT faster
template<class lcp_t, class bwt_t>
vector<uint32_t> ct2(lcp_t& lcp, bwt_t& bwt, uint64_t largest_symbol)
{
	rmq_succinct_sct<> rmq(&lcp);
	int64_t t1 = 5;
	uint64_t t2 = 10;
	vector<uint64_t> cache(t1, 0);
	vector<uint32_t> ct(bwt.size(), 0);
	vector<uint64_t> last(largest_symbol+1, 0);
	for(uint64_t i=0; i<ct.size(); ++i)
	{
		uint64_t c = bwt[i];
		if(lcp[i] < t1)
		{
			cache[lcp[i]] = i;
		}
		if(last[c] != 0)
		{
			uint64_t j = last[c], start = last[c];
			if(i-start<t2)
			{
				for(uint64_t idx=start+1; idx<=i; ++idx)
				{
					if(lcp[idx] < lcp[j])
					{
						j = idx;
					}
				}
			}
			else
			{
				uint64_t k=0;
				while(cache[k]<start and k<cache.size())
				{
					++k;
				}
				if(k<cache.size())
				{
					j = cache[k];
				}
				else
				{
					j = rmq(start, i);
				}
			}
			++ct[j];
		}
		last[c] = i+1;
	}
	for(uint64_t i=1; i<ct.size(); ++i)
	{
		ct[i] += ct[i-1];
	}
	return ct;
}

struct four_tupel
{
	int64_t lcp;
	int64_t lb;
	int64_t rb;
	int64_t lindizes;
	four_tupel(int64_t _lcp, int64_t _lb, int64_t _rb=-1, int64_t _lindizes=0) : lcp(_lcp), lb(_lb), rb(_rb), lindizes(_lindizes) {}
};

void process(int64_t p, int64_t q, four_tupel& interval, vector<uint32_t>& ct)
{
	if(interval.lcp and (interval.rb-interval.lb+1-(ct[interval.rb]-ct[interval.lb])) >=p and (interval.lindizes+1) >= q)
	{
		++repeats_count;
	}
}

template<class lcp_t, class ct_t>
void pq_context_repeats(int64_t p, int64_t q, lcp_t& lcp, ct_t& ct)
{
	stack<four_tupel> s;
	s.push(four_tupel(-1, 0));
	for(uint64_t k=1, lb=0; k<lcp.size(); ++k)
	{
		lb = k-1;
		while(lcp[k] < s.top().lcp)
		{
			s.top().rb = k-1;
			four_tupel last_interval = s.top(); s.pop();
			process(p, q, last_interval, ct);
			lb = last_interval.lb;
		}
		if(lcp[k] > s.top().lcp)
		{
			s.push(four_tupel(lcp[k], lb, -1, 1));
		}
		else
		{
			s.top().lindizes += 1;
		}
	}
}

int main (int argc, char** argv)
{
	// Check Parameters
	if(argc!=4)
	{
		cerr << "Program takes path of input, p and q." << endl;
		return 1;
	}

	// Read Text, as 1-Byte-Char File
	vector<uint8_t> text;
	ifstream input(argv[1], std::ios::in|std::ios::binary|std::ios::ate);
	uint64_t n = input.tellg();
	input.seekg(0);
	text.resize(n+1);
	input.read((char*)text.data(), n);

	// Create Suffix Array with DivSufSort
	vector<uint32_t> sa(text.size());
	divsufsort((const sauchar_t*)text.data(), (int32_t*)(sa.data()+1), n);

	// Append Seninal
	text[n] = 0;
	sa[0] = n;

	// Read p and q
	int64_t p=atoi(argv[2]);
	int64_t q=atoi(argv[3]);

	// Create LCP-Array with phi
	vector<int32_t> lcp(sa.size()+1, 0);
	PHI(text, sa, lcp);

	// Search largest_symbol
	uint64_t largest_symbol = 0;
	for(uint64_t i=0; i<text.size(); ++i)
	{
		if(largest_symbol < text[i])
		{
			largest_symbol = text[i];
		}
	}

	// Create BWT
	vector<uint8_t> bwt(sa.size());
	for(uint64_t i=0; i<bwt.size(); ++i)
	{
		if(sa[i])
		{
			bwt[i] = text[sa[i]-1];
		}
		else
		{
			bwt[i] = text[text.size()-1];
		}
	}
	// SA and Text are no longer needed
	clear(sa);
	clear(text);

	// Create ct
#ifdef CT1
	vector<uint32_t> ct = ct1(lcp, bwt, largest_symbol);
#endif
#ifdef CT2
	vector<uint32_t> ct = ct2(lcp, bwt, largest_symbol);
#endif
	clear(bwt);

	// calculate p,q-context repeats
	pq_context_repeats(p, q, lcp, ct);
	cout << repeats_count << "\t<" << p << "," << q << ">-context-diverse-repeats in '" << argv[1] << "'."  << endl;

	return 0;
}
